#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 09:57:01 2020

@author: RMS671214
"""

from numpy import datetime64 as dt64
from .rmp_dates import generate_dates as gen_dates, \
    frequencies as fre,  day_count_factor as day_cf
from .conventions import start_basis, frequencies
from .rmp_curves import interpolation, calc_shortrate_from_df, \
calc_df_from_shortrate, calc_fwd_df, shift_curve, discount_factor_from_zspread
from .discount_curve import discount_factor_from_ytm as ytm_df, \
    discount_factor_from_ytm_using_structures as ytm_df_struct, \
    discount_factor_gen as df_gen
from collections import deque
import numpy as np
import math


def fixbond(bond):

    # create structure for the bond
    structure = fixbond_structures(bond)
    risks = fixbond_value(bond["value_date"], structure, bond["ytm"],
                          bond["day_count"], bond["frequency"],
                          bond["business_day"])

    return {"structure": structure, "risks": risks}


def fixbond_value(value_date, structures, yld, day_count, frequency,
                  business_day="No Adjustment"):
    const = 0.01
    try:
        ytm = float(yld)
        ytm1 = ytm + const
        ytm2 = ytm1 + const
    except Exception:
        return None
    newstructures = [dict(x) for x in structures]
    maturity = newstructures[-1]["end_date"]

    # Curves to be used in calculation of duration, convexity and pvbp01
    df_curve = ytm_df_struct(value_date, newstructures, day_count, frequency,
                             business_day, ytm)
    x_axis = [x["times"] for x in df_curve]
    y_axis = [x["df"] for x in df_curve]
    ifunc = interpolation(x_axis, y_axis, float(1/366), is_function=True)

    df_curve1 = ytm_df_struct(value_date, newstructures, day_count, frequency,
                              business_day, ytm1)
    x_axis1 = [x["times"] for x in df_curve1]
    y_axis1 = [x["df"] for x in df_curve1]
    ifunc1 = interpolation(x_axis1, y_axis1, float(1/366), is_function=True)

    df_curve2 = ytm_df_struct(value_date, newstructures, day_count, frequency,
                              business_day, ytm2)
    x_axis2 = [x["times"] for x in df_curve2]
    y_axis2 = [x["df"] for x in df_curve2]
    ifunc2 = interpolation(x_axis2, y_axis2, float(1/366), is_function=True)

    # covert data into list
    dates = [{"start_date": x["start_date"], "end_date": x["end_date"]}
             for x in structures if value_date < dt64(x["end_date"])]
    cfs = np.asarray([x["cash_flow"] for x in structures
                      if value_date < dt64(x["end_date"])])

    times = np.asarray([day_cf("Actual/365", value_date, x["end_date"])
                        for x in dates])
    # interpolating discount factors
    dfs = [ifunc(x) for x in times]
    # calculating the present values
    pvs = cfs * dfs
    weighted_pvs = pvs * times
    value = np.sum(pvs)
    weighted_value = np.sum(weighted_pvs)

    # interpolating 2nd discount factors
    dfs1 = [ifunc1(x) for x in times]
    # calculating the 2nd present values
    pvs1 = cfs * dfs1
    value1 = np.sum(pvs1)

    # interpolating 3rd discount factors
    dfs2 = [ifunc2(x) for x in times]
    # calculating the 3rd present values
    pvs2 = cfs * dfs2
    value2 = np.sum(pvs2)

    mac_dur = weighted_value/value
    compound = 12 / frequencies[frequency]
    mod_dur = mac_dur / pow((1 + ytm / (compound * 100)), compound)
    pvbp01 = value1 - value

    # 1st derivative at ytm using forward differential
    der0 = (value1 - value) / (ytm1 - ytm)
    # 1st derivative at ytm1 using forward differential
    der1 = (value2 - value1) / (ytm2 - ytm1)
    # 2nd derivative at ytm using forward differential
    conv = (der1 - der0) / (ytm2-ytm1)

    return {"macaulay_duration": mac_dur, "modified_duration": mod_dur,
            "pvbp01": pvbp01, "convexity": conv, "value": value}


def fixbond_structures(bond):
    """
    Generate bonds coupon structures.
            Parameters:
                bonds: a dictionary with the following keys - value_date,
                business_day, issue_date, value_date, maturity,
                frequency, day_count, date_generation, face_value, coupon,
                ytm and type.
            Returns:
                a dictionary with the following keys - date, dcf, time,
                days, df, rate
    """
    results = list(_fixbond_gen_structure(bond))

    return results


def _fixbond_gen_structure(bond):

    dates = _dates_gen_structure(bond)
    dates = deque(dates)

    noofcpns = len(dates)
    for no in range(noofcpns):
        structure = {}
        structure["cpn_dcf"] = day_cf(bond["day_count"],
                                      dates[no]["start_date"],
                                      dates[no]["end_date"],
                                      bondmat_date=bond["maturity"],
                                      next_coupon_date=dates[no]["end_date"],
                                      business_day=bond["business_day"],
                                      Frequency=12/frequencies[bond["frequency"]])
        structure["coupon"] = bond["coupon"]
        structure["face_value"] = bond["face_value"]
        structure["coupon_interest"] = (structure["cpn_dcf"] *
                                        structure["coupon"] *
                                        structure["face_value"] / 100)
        if no != noofcpns - 1:
            structure["fv_flow"] = 0
        else:
            structure["fv_flow"] = structure["face_value"]
        structure["cash_flow"] = (structure["coupon_interest"] +
                                  structure["fv_flow"])

        dates[no].update(structure) # merging the dictionary

    if (bond.get("ytm") and bond["value_date"]):
        fixbond_value(bond["value_date"], dates, bond["ytm"], bond["day_count"],
                      bond["frequency"])

    return dates


def date_structures(bonds):
    """
    Generate coupon structures.

            Parameters:
                bonds: a dictionary with the following keys - value_date,
                business_day, issue_date, value_date, maturity,
                frequency, day_count, date_generation, face_value, coupon,
                ytm and type.

            Returns:
                a dictionary with the following keys - date, dcf, time,
                days, df, rate
    """
    if isinstance(bonds, dict):
        results = _dates_gen_structure(bonds)

    elif isinstance(bonds, list):
        results = []
        for bond in bonds:
            if isinstance(bond, dict):
                results.append(_dates_gen_structure(bond))
    return results


def _dates_gen_structure(bond):
    """
    Generate date structures. Function can be used for all coupon bearing
    products with bullet principal repayment

            Parameters:
                bonds: a dictionary with the following keys - value_date,
                business_day, issue_date, value_date, maturity,
                frequency, day_count, date_generation.

            Returns:
                a dictionary or an array of dictionaries with the following
                keys - "start_date" and "end_date"
    """
    bus_day = None
    if bond['business_day'] == 'NULL':
        bus_day = 'No Adjustment'
    else:
        bus_day = bond['business_day']


    start_date = bond.get('issue_date')
    if start_date is not None:
        start_date = dt64(start_date, 'D')
    value_date = bond.get('value_date')

    if value_date is not None:
        value_date = dt64(value_date, 'D')
    if start_date is not None and value_date is not None:
        use_date = start_date
    elif start_date is not None:
        use_date = start_date
    elif value_date is not None:
        use_date = value_date
    else:
        raise Exception('Both value_date and issue_date do not have any value')
    # dates is a deque
    dates = gen_dates(use_date, bond['maturity'], issueDate=start_date,
                      frequency=fre[bond['frequency']],
                      business_day=bus_day, method=bond['date_generation'])

    start_dates = deque(dates)
    end_dates = deque(dates)
    start_dates.pop()
    end_dates.popleft()
    noofcpns = len(start_dates)
    structures = deque()
    for no in range(noofcpns):
        structure = {"start_date": start_dates[no], "end_date": end_dates[no]}
        structures.append(structure)

    newstructure = list(map(lambda sdate, edate: {"start_date": sdate,
                                                  "end_date": edate},
                            start_dates, end_dates))

    return newstructure


def floatbond(bond, ref_curve, val_curve=None):
    """
    Generate bonds coupon structures.
            Parameters:
                bonds: a dictionary with the following keys - value_date,
                business_day, issue_date, value_date, maturity,
                frequency, day_count, date_generation, face_value, coupon,
                margin.
            Returns:
                a dictionary with the following keys - date, dcf, time,
                days, df, rate
    """
    # create structure for the bond
    structure = list(_floatbond_gen_structure(bond))

    risks = floatbond_value(bond["value_date"], structure, bond["spread"],
                            bond["day_count"], bond["frequency"],
                            ref_curve=ref_curve, val_curve=val_curve)

    return {"structure": structure, "risks": risks}


def _floatbond_gen_structure(bond, holidays=[]):

    # generate the face value and coupons
    dates = _dates_gen_structure(bond)
    dates = deque(dates)

    bdc = np.busdaycalendar(weekmask='1111100', holidays=holidays)
    for date in dates:
        offset = -start_basis[bond["fixing_basis"]]
        date["fixing_date"] = np.busday_offset(date["start_date"], offset,
                                               roll='backward', busdaycal=bdc)
        date["face_value"] = bond["face_value"]
        date["margin"] = bond["margin"]
        date["cpn_dcf"] = day_cf(bond["day_count"],
                                 date["start_date"],
                                 date["end_date"],
                                 bondmat_date=bond["maturity"],
                                 next_coupon_date=date["end_date"],
                                 business_day=bond["business_day"],
                                 Frequency=bond["frequency"])
        # fixing date is a forward date
        if date["fixing_date"] > bond["value_date"]:
            date["is_fixed"] = False
            date["accrued"] = 0

        # fixing date is for the current coupon period
        elif (date["fixing_date"] <= bond["value_date"] and
              date["end_date"] > bond["value_date"]):
            date["is_fixed"] = True
            date["coupon"] = bond["current_coupon"]
            date["coupon_interest"] = (date["cpn_dcf"] *
                                       date["coupon"] *
                                       date["face_value"] / 100)
            if bond["value_date"] > date["start_date"]:
                acc_dcf = day_cf(bond["day_count"],
                                 date["start_date"],
                                 bond["value_date"],
                                 bondmat_date=bond["maturity"],
                                 next_coupon_date=date["end_date"],
                                 business_day=bond["business_day"],
                                 Frequency=bond["frequency"])
                date["accrued"] = acc_dcf / date["cpn_dcf"] * date["coupon_interest"]
            else:
                date["accrued"] = 0

        else:
            date["is_fixed"] = True
            date["accrued"] = 0
        date["fv_flow"] = 0
    dates[-1]["fv_flow"] = bond["face_value"]

    return dates


def floatbond_value(value_date, structures, spread, day_count, frequency,
                    ref_curve, val_curve=None):
    if val_curve is None:
        risks = _float_val_use_zspread(value_date, structures, spread,
                                       day_count, frequency, ref_curve)
    else:
        risks = _float_val_use_curve(value_date, structures, day_count,
                                     frequency, ref_curve, val_curve)

    return risks


def _float_val_use_zspread(value_date, structures, zspread, day_count,
                           frequency, ref_curve):

    # copy only active structure period
    data = [dict(x) for x in structures if value_date < x["end_date"]]
    # covert data into list
    times = np.asarray([day_cf("Actual/365", value_date, x["end_date"])
                        for x in data])

    # Original Curves
    ref_df_curve = df_gen(ref_curve, return_type="times")
    val_df_curve = discount_factor_from_zspread(value_date, structures,
                                                day_count, frequency,
                                                ref_df_curve, zspread)

    # Shift the original curves by 1 bp
    ref_curve1 = dict(ref_curve)
    ref_curve1["rates"] = shift_curve(ref_curve1["rates"])
    ref_df_curve1 = df_gen(ref_curve1, return_type="times")
    val_df_curve1 = discount_factor_from_zspread(value_date, structures,
                                                 day_count, frequency,
                                                 ref_df_curve1, zspread)

    # Shift the curves by another 1bp
    ref_curve2 = dict(ref_curve1)
    ref_curve2["rates"] = shift_curve(ref_curve2["rates"])
    ref_df_curve2 = df_gen(ref_curve2, return_type="times")
    val_df_curve2 = discount_factor_from_zspread(value_date, structures,
                                                 day_count, frequency,
                                                 ref_df_curve2, zspread)

    pvs0 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve, val_df_curve)
    value = np.sum(pvs0)

    pvs1 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve1, val_df_curve1)
    value1 = np.sum(pvs1)

    pvs2 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve2, val_df_curve2)
    value2 = np.sum(pvs2)
    pvbp01 = value1 - value
    duration = pvbp01 / 0.01
    duration2 = (value2 - value1) / 0.01
    convexity = (duration2 - duration) / 0.01

    return {"duration": duration, "pvbp01": pvbp01, "convexity": convexity,
            "value": value}


def _float_val_use_curve(value_date, structures, day_count, frequency,
                         ref_curve, val_curve):

    # copy only active structure period
    data = [dict(x) for x in structures if value_date < x["end_date"]]
    # covert data into list
    times = np.asarray([day_cf("Actual/365", value_date, x["end_date"])
                        for x in data])

    # Original Curves
    ref_df_curve = df_gen(ref_curve, return_type="times")
    val_df_curve = df_gen(val_curve, return_type="times")

    # Shift the original curves by 1 bp
    ref_curve1 = dict(ref_curve)
    ref_curve1["rates"] = shift_curve(ref_curve1["rates"])
    ref_df_curve1 = df_gen(ref_curve1, return_type="times")

    val_curve1 = dict(val_curve)
    val_curve1["rates"] = shift_curve(val_curve1["rates"])
    val_df_curve1 = df_gen(val_curve1, return_type="times")

    # Shift the curves by another 1bp
    ref_curve2 = dict(ref_curve1)
    ref_curve2["rates"] = shift_curve(ref_curve2["rates"])
    ref_df_curve2 = df_gen(ref_curve2, return_type="times")

    val_curve2 = dict(val_curve1)
    val_curve2["rates"] = shift_curve(val_curve2["rates"])
    val_df_curve2 = df_gen(val_curve2, return_type="times")

    pvs0 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve, val_df_curve)
    value = np.sum(pvs0)

    pvs1 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve1, val_df_curve1)
    value1 = np.sum(pvs1)

    pvs2 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve2, val_df_curve2)
    value2 = np.sum(pvs2)
    pvbp01 = value1 - value
    duration = pvbp01 / 0.01
    duration2 = (value2 - value1) / 0.01
    convexity = (duration2 - duration) / 0.01

    return {"duration": duration, "pvbp01": pvbp01, "convexity": convexity,
            "value": value}


def _float_recalc_loop(value_date, data, day_count, frequency,
                       times,  df_curve, val_df_curve):

    x_axis = [x["times"] for x in df_curve]
    y_axis = [x["df"] for x in df_curve]
    ifunc = interpolation(x_axis, y_axis, float(1/366), is_function=True)

    x_val = [x["times"] for x in val_df_curve]
    y_val = [x["df"] for x in val_df_curve]
    ifunc_val = interpolation(x_val, y_val, float(1/366), is_function=True)
    val_dis_curve = [ifunc_val(x) for x in times]

    for datum in data:
        end_time = day_cf("Actual/365", value_date, datum["end_date"])

        if not datum["is_fixed"]:  # coupon has not been fixed
            start_time = day_cf("Actual/365", value_date,
                                datum["start_date"])

            fwd_df = calc_fwd_df(start_time, end_time, ifunc=ifunc)
            fwd_rate = calc_shortrate_from_df(datum["start_date"],
                                              datum["end_date"],
                                              fwd_df, day_count)
            datum["coupon"] = fwd_rate + datum["margin"]
            datum["coupon_interest"] = (datum["coupon"] * datum["cpn_dcf"] *
                                        datum["face_value"] / 100)

        if datum.get("coupon_interest"):
            datum["cash_flow"] = datum["coupon_interest"] + datum["fv_flow"]
        else:
            datum["cash_flow"] = datum["fv_flow"]

    cash_flows = np.asarray([datum["cash_flow"] for datum in data])
    pvs = cash_flows * val_dis_curve

    return pvs




def fixleg(data, val_curve=None):
    structures = _fixleg_gen_structure(data)
    if not data["principal_exchange"]:
        for structure in structures:
            structure["cash_flow"] = structure["coupon_interest"]

    if val_curve is not None:
        risks = fixleg_value(data["value_date"], structures, data["day_count"],
                             data["frequency"], val_curve)
    else:
        risks = None

    return {"structure": structures, "risks": risks}


def _fixleg_gen_structure(data):

    dates = _dates_gen_structure(data)
    dates = deque(dates)

    noofcpns = len(dates)
    for no in range(noofcpns):
        structure = {}
        structure["cpn_dcf"] = day_cf(data["day_count"],
                                      dates[no]["start_date"],
                                      dates[no]["end_date"],
                                      bondmat_date=data["maturity"],
                                      next_coupon_date=dates[no]["end_date"],
                                      business_day=data["business_day"],
                                      Frequency=12/frequencies[data["frequency"]])
        structure["coupon"] = data["coupon"]
        structure["face_value"] = data["face_value"]
        structure["coupon_interest"] = (structure["cpn_dcf"] *
                                        structure["coupon"] *
                                        structure["face_value"] / 100)
        if no != noofcpns - 1:
            structure["fv_flow"] = 0
        else:
            structure["fv_flow"] = structure["face_value"]
        structure["cash_flow"] = (structure["coupon_interest"] +
                                  structure["fv_flow"])

        dates[no].update(structure) # merging the dictionary


    return dates


def fixleg_value(value_date, structures, day_count, frequency, val_curve):

    # copy only active structure period
    data = [dict(x) for x in structures if value_date < x["end_date"]]
    # covert data into list
    times = np.asarray([day_cf("Actual/365", value_date, x["end_date"])
                        for x in data])

    # Original Curves
    val_df_curve = df_gen(val_curve, return_type="times")

    # Shift the original curves by 1 bp
    val_curve1 = dict(val_curve)
    val_curve1["rates"] = shift_curve(val_curve1["rates"])
    val_df_curve1 = df_gen(val_curve1, return_type="times")

    # Shift the curves by another 1bp
    val_curve2 = dict(val_curve1)
    val_curve2["rates"] = shift_curve(val_curve2["rates"])
    val_df_curve2 = df_gen(val_curve2, return_type="times")

    pvs0 = _fixleg_recalc_loop(value_date, data, day_count, frequency,
                               times, val_df_curve)
    value = np.sum(pvs0)

    pvs1 = _fixleg_recalc_loop(value_date, data, day_count, frequency,
                               times, val_df_curve1)
    value1 = np.sum(pvs1)

    pvs2 = _fixleg_recalc_loop(value_date, data, day_count, frequency,
                               times, val_df_curve2)
    value2 = np.sum(pvs2)
    pvbp01 = value1 - value
    duration = pvbp01 / 0.01
    duration2 = (value2 - value1) / 0.01
    convexity = (duration2 - duration) / 0.01

    return {"duration": duration, "pvbp01": pvbp01, "convexity": convexity,
            "value": value}


def _fixleg_recalc_loop(value_date, data, day_count, frequency,
                        times, val_df_curve):

    x_val = [x["times"] for x in val_df_curve]
    y_val = [x["df"] for x in val_df_curve]
    ifunc_val = interpolation(x_val, y_val, float(1/366), is_function=True)
    val_dis_curve = [ifunc_val(x) for x in times]

    cash_flows = np.asarray([datum["cash_flow"] for datum in data])
    pvs = cash_flows * val_dis_curve

    return pvs


def floatleg(data, ref_curve, val_curve=None):
    structures = _floatleg_gen_structure(data)
    if not data["principal_exchange"]:
        for structure in structures:
            structure["fv_flow"] = 0

    if val_curve is not None:
        risks = _floatleg_value(data["value_date"], structures,
                                data["day_count"], data["frequency"],
                                ref_curve, val_curve)
    else:
        risks = None

    return {"structure": structures, "risks": risks}


def _floatleg_gen_structure(data, holidays=[]):

    # generate the face value and coupons
    dates = _dates_gen_structure(data)
    dates = deque(dates)

    bdc = np.busdaycalendar(weekmask='1111100', holidays=holidays)
    for date in dates:
        offset = -start_basis[data["fixing_basis"]]
        date["fixing_date"] = np.busday_offset(date["start_date"], offset,
                                               roll='backward', busdaycal=bdc)
        date["face_value"] = data["face_value"]
        date["margin"] = data["margin"]
        date["cpn_dcf"] = day_cf(data["day_count"],
                                 date["start_date"],
                                 date["end_date"],
                                 bondmat_date=data["maturity"],
                                 next_coupon_date=date["end_date"],
                                 business_day=data["business_day"],
                                 Frequency=data["frequency"])
        # fixing date is a forward date
        if date["fixing_date"] > data["value_date"]:
            date["is_fixed"] = False
            date["accrued"] = 0

        # fixing date is for the current coupon period
        elif (date["fixing_date"] <= data["value_date"] and
              date["end_date"] > data["value_date"]):
            date["is_fixed"] = True
            date["coupon"] = data["current_coupon"]
            date["coupon_interest"] = (date["cpn_dcf"] *
                                       date["coupon"] *
                                       date["face_value"] / 100)
            if data["value_date"] > date["start_date"]:
                acc_dcf = day_cf(data["day_count"],
                                 date["start_date"],
                                 data["value_date"],
                                 bondmat_date=data["maturity"],
                                 next_coupon_date=date["end_date"],
                                 business_day=data["business_day"],
                                 Frequency=data["frequency"])
                date["accrued"] = (acc_dcf / date["cpn_dcf"] *
                                   date["coupon_interest"])
            else:
                date["accrued"] = 0

        else:
            date["is_fixed"] = True
            date["accrued"] = 0
        date["fv_flow"] = 0
    dates[-1]["fv_flow"] = data["face_value"]

    return dates


def _floatleg_value(value_date, structures, day_count, frequency,
                    ref_curve, val_curve):

    # copy only active structure period
    data = [dict(x) for x in structures if value_date < x["end_date"]]
    # covert data into list
    times = np.asarray([day_cf("Actual/365", value_date, x["end_date"])
                        for x in data])

    # Original Curves
    ref_df_curve = df_gen(ref_curve, return_type="times")
    val_df_curve = df_gen(val_curve, return_type="times")

    # Shift the original curves by 1 bp
    ref_curve1 = dict(ref_curve)
    ref_curve1["rates"] = shift_curve(ref_curve1["rates"])
    ref_df_curve1 = df_gen(ref_curve1, return_type="times")

    val_curve1 = dict(val_curve)
    val_curve1["rates"] = shift_curve(val_curve1["rates"])
    val_df_curve1 = df_gen(val_curve1, return_type="times")

    # Shift the curves by another 1bp
    ref_curve2 = dict(ref_curve1)
    ref_curve2["rates"] = shift_curve(ref_curve2["rates"])
    ref_df_curve2 = df_gen(ref_curve2, return_type="times")

    val_curve2 = dict(val_curve1)
    val_curve2["rates"] = shift_curve(val_curve2["rates"])
    val_df_curve2 = df_gen(val_curve2, return_type="times")

    pvs0 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve, val_df_curve)
    value = np.sum(pvs0)

    pvs1 = _float_recalc_loop(value_date, data, day_count, frequency,
                              times, ref_df_curve1, val_df_curve1)
    value1 = np.sum(pvs1)

    pvs2 = _floatleg_recalc_loop(value_date, data, day_count, frequency,
                                 times, ref_df_curve2, val_df_curve2)
    value2 = np.sum(pvs2)
    pvbp01 = value1 - value
    duration = pvbp01 / 0.01
    duration2 = (value2 - value1) / 0.01
    convexity = (duration2 - duration) / 0.01

    return {"duration": duration, "pvbp01": pvbp01, "convexity": convexity,
            "value": value}


def _floatleg_recalc_loop(value_date, data, day_count, frequency,
                          times,  df_curve, val_df_curve):

    x_axis = [x["times"] for x in df_curve]
    y_axis = [x["df"] for x in df_curve]
    ifunc = interpolation(x_axis, y_axis, float(1/366), is_function=True)

    x_val = [x["times"] for x in val_df_curve]
    y_val = [x["df"] for x in val_df_curve]
    ifunc_val = interpolation(x_val, y_val, float(1/366), is_function=True)
    val_dis_curve = [ifunc_val(x) for x in times]

    for datum in data:
        end_time = day_cf("Actual/365", value_date, datum["end_date"])

        if not datum["is_fixed"]:  # coupon has not been fixed
            start_time = day_cf("Actual/365", value_date,
                                datum["start_date"])

            fwd_df = calc_fwd_df(start_time, end_time, ifunc=ifunc)
            fwd_rate = calc_shortrate_from_df(datum["start_date"],
                                              datum["end_date"],
                                              fwd_df, day_count)
            datum["coupon"] = fwd_rate + datum["margin"]
            datum["coupon_interest"] = (datum["coupon"] * datum["cpn_dcf"] *
                                        datum["face_value"] / 100)

        if datum.get("coupon_interest"):
            datum["cash_flow"] = datum["coupon_interest"] + datum["fv_flow"]
        else:
            datum["cash_flow"] = datum["fv_flow"]

    cash_flows = np.asarray([datum["cash_flow"] for datum in data])
    pvs = cash_flows * val_dis_curve

    return pvs


def swaps(fxleg, flleg, fxlval_curve=None, flref_curve=None,
          fllval_curve=None, fxrate=1.00):

    flleg = floatleg(flleg, ref_curve=flref_curve, val_curve=fllval_curve)
    fxleg = fixleg(fxleg, val_curve=fxlval_curve)
    fxleg_risks = fxleg["risks"]
    flleg_risks = flleg["risks"]

    swapvalue = fxleg_risks["value"] * fxrate - flleg_risks["value"]
    swappvbp01 = fxleg_risks["pvbp01"] * fxrate - flleg_risks["pvbp01"]

    return {"value": swapvalue, "pvbp01": swappvbp01}

def loans_structures(loans):
    """
    Generate loan schedule.

            Parameters:
                loans: a dictionary with the following keys - value_date,
                business_day, start_date, value_date, maturity,
                frequency, day_count, date_generation, face_value, rate,
                rate_type and rate_compounding.

            Returns:
                a dictionary with the following keys - date, dcf, time,
                days, df, rate
    """

    if isinstance(loans, dict):
        results = _loan_gen_structure(loans)

    elif isinstance(loans, list):
        results = []
        for loan in loans:
            if isinstance(loan, dict):
                results.append(_fixbond_gen_structure(loan))

    return results


def _loan_gen_structure(loan):
    value_date = loan.get("value_date")
    structures = _dates_gen_structure(loan)
    if loan["rate_type"] == "fixed":
        for structure in structures:
            pass
    elif loan["rate_type"] == float:
        pass

    return structures


def calc_customfix_structures(structures, day_count, frequency, business_day):
    """
    Calculate all the values of the custome structure.

            Parameters:
                value_date: numpy.datetime64.
                structures: a list of dictionaries with the following keys: start_date, end_date, face_value, coupon, fv_flow.
                day_count: str
                frequeny: str
            Returns:
                a dictionary with the following keys - value_date, start_date, end_date, face_value, coupon, coupon_interest, cash_flow and cpn_dcf.
    """

    for structure in structures:
        structure["cpn_dcf"] = day_cf(day_count,
                                      structure["start_date"],
                                      structure["end_date"],
                                      bondmat_date=structure["end_date"],
                                      next_coupon_date=structure["end_date"],
                                      business_day=business_day,
                                      Frequency=frequency)

        structure["coupon_interest"] = (structure["cpn_dcf"] *
                                        structure["coupon"] *
                                        structure["face_value"] / 100)
        if structure.get("fv_flow"):
            structure["cash_flow"] = (structure["coupon_interest"] +
                                            structure["fv_flow"])
        else:
            structure["fv_flow"] = 0.00
            structure["cash_flow"] = (structure["coupon_interest"])

    return structures


def value_customfix_structures(value_date, structures, day_count, frequency,
                               dis_curve):
    """
    Calculate all the values of the custome structure.

            Parameters:
                value_date: numpy.datetime64.
                structures: a list of dictionaries with the following keys:
                    start_date, end_date, face_value, coupon and fv_flow.
                day_count: str
                frequeny: str
                dis_curve: list of ditionaries with the following keys: time
                and df
            Returns:
                a dictionary with the following keys - value_date, start_date,
                end_date, face_value, coupon, coupon_interest, cash_flow and
                cpn_dcf.
    """
    newstructures = [dict(x) for x in structures]

    if isinstance(dis_curve, list):
        xaxis = [x["times"] for x in dis_curve]
        yaxis = [x["df"] for x in dis_curve]
        ifunc = interpolation(xaxis, yaxis, 1, model='chip', is_function=True)

    for structure in newstructures:
        temp_structure = {}
        temp_structure["cpn_dcf"] = day_cf(day_count,
                                      structure["start_date"],
                                      structure["end_date"],
                                      bondmat_date=structure["end_date"],
                                      next_coupon_date=structure["end_date"],
                                      Frequency=frequency)

        temp_structure["coupon_interest"] = (temp_structure["cpn_dcf"] *
                                        structure["coupon"] *
                                        structure["face_value"] / 100)
        if structure.get("fv_flow"):
            temp_structure["cash_flow"] = (temp_structure["coupon_interest"] +
                                            structure["fv_flow"])
        else:
            structure["fv_flow"] = 0.00
            temp_structure["cash_flow"] = (temp_structure["coupon_interest"])
        temp_structure["time"] = float(day_cf(day_count,
                                      value_date,
                                      structure["end_date"]))
        structure.update(temp_structure) # merging the dictionary

    # Discount curve was provided
    if isinstance(dis_curve, list):
        for structure in newstructures:
            if structure["time"] > 0:
                structure["df"] = float(ifunc(structure["time"]))
                structure["pv"] = structure["df"] * structure["cash_flow"]
            else:
                structure["df"] = 0
                structure["pv"] = 0

    # yield to maturity was provided instead
    elif isinstance(dis_curve, float):
        newstructures = fixbond_value(value_date, newstructures, dis_curve,
                                      day_count, frequency)

    return newstructures


def _calc_customfix_risks(value_date, structures, day_count, frequency,
                               dis_curve):
    data = list(structures)
    mylist = []
    for datum in data:
        mydatum = dict(datum)
        mylist.append

    if isinstance(dis_curve, list):
        pass
    elif isinstance(dis_curve, float):
        ytm = dis_curve
    # Calculate the PV weight
    maturity = mylist[-1]["end_date"]
    for datum in data:
            datum["time"] = day_cf("Actual/365",
                                   value_date,
                                   datum["end_date"],
                                   bondmat_date=maturity,
                                   next_coupon_date=datum["end_date"],
                                   business_day="No Adjustment",
                                   Frequency=frequency)
            datum["cf_weight"] = datum["pv"] * datum['time']
            datum["period_df1"] = 1 / (1 + datum["ytm_dcf"] *
                                       (bond["ytm"] + 0.01) / 100)

    mac_dur = 0
    val = 0
    for datum in data:
        if datum["end date"] > bond["value_date"]:
            mac_dur += datum["cf_weight"]
            val += datum["pv"]
    mac_dur = mac_dur / val
    mod_dur = mac_dur(1 + bond["ytm"] /
                      (12 / frequencies[bond["frequency"]]))


def create_structures_from_dates(dates, coupons, face_values, fv_flows):
    """
    Create structures from date.

            Parameters:
                dates: list of dict with the following keys - start_date and
                end_date
                coupons: list of float.
                face_values: list of float
                fv_flows: list of float.

            Returns:
                list of dictionaries with the following keys - start_date, end_date, coupon, face_value, fv_flow
    """

    structures = list(map( lambda date, coupon, face_value, fv_flow: {"start_date": date["start_date"], "end_date": date["end_date"], "coupon": coupon, "face_value": face_value, "fv_flow": fv_flow}, dates, coupons, face_values, fv_flows))

    return structures



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 17:29:52 2020

@author: RMS671214
"""
from numpy import datetime64 as dt64
from .conventions import *
from .rmp_dates import day_count_factor as day_cf
from .rmp_bhelp import *



class Bond:
    def __init__(self):
        self.__risk_stats = {}

    @property
    def issue_date(self):
        return self.__issue_date

    @issue_date.setter
    def issue_date(self, issue_date):
        if issue_date is None:
            self.__issue_date = None
        else:
            try:
                self.__issue_date = dt64(issue_date)
            except:
                raise ValueError('Bond class error: issue_date is not a date')

    @property
    def last_coupon_date(self):
        return self.__last_coupon_date

    @last_coupon_date.setter
    def last_coupon_date(self, last_coupon_date):
        if last_coupon_date is None:
            self.__last_coupon_date = None
        else:
            try:
                self.__last_coupon_date = dt64(last_coupon_date)
            except:
                raise ValueError('Bond class error: last_coupon_date is '
                                 + 'not a date')

    @property
    def value_date(self):
        return self.__value_date

    @value_date.setter
    def value_date(self, value_date):
        if value_date is None:
            self.__value_date = None
        else:
            try:
                self.__value_date = dt64(value_date)

            except:
                raise ValueError('Bond class error: ' + value_date + ' passed'
                                 + ' for value date is not a date')


    @property
    def maturity(self):
        return self.__maturity

    @maturity.setter
    def maturity(self, maturity):
        if maturity is None:
            self.__maturity = maturity
        else:
            try:
                self.__maturity = dt64(maturity)
            except:
                raise ValueError('Bond class error: ' + maturity + ' passed'
                                 + ' for date argument is not a date')

    @property
    def day_count(self):
        return self.__day_count

    @day_count.setter
    def day_count(self, day_count):
        if day_count is None or day_count in day_counts:
            self.__day_count = day_count
        else:
            raise ValueError('Bond class error: ' + day_count + ' is not in'
                             + ' the list of conventions')


    @property
    def business_day(self):
        return self.__business_day

    @business_day.setter
    def business_day(self, business_day):
        if business_day is None or business_day in business_days:
            self.__business_day = business_day
        else:
            raise ValueError('Bond class error: ' + business_day + ' is not in'
                             + ' the list of conventions')


    @property
    def date_gen(self):
        return self.__date_gen

    @date_gen.setter
    def date_gen(self, date_gen):
        if date_gen is None or date_gen in date_gen_method:
            self.__date_gen = date_gen
        else:
            raise ValueError('Bond class error: ' + date_gen + ' is not'
                             + ' in the list of methods')


    @property
    def frequency(self):
        return self.__frequency

    @frequency.setter
    def frequency(self, frequency):
        if frequency is None or frequency in frequencies:
            self.__frequency = frequency
        else:
            raise ValueError('Bond class error: ' + frequency + ' is not in'
                             + ' the list of frequencies')


class FixBond(Bond):

    def __init__(self):
        super().__init__()
        self.__risk_stats = {}

    @property
    def coupon(self):
        return self.__coupon

    @coupon.setter
    def coupon(self, coupon):
        if coupon is None:
            self.__coupon = coupon
        else:
            try:
                self.__coupon = float(coupon)
            except:
                raise ValueError('Bond class error: ' + coupon + ' is not'
                                 + ' a number')


    @property
    def accrued_interest(self):
        return self.__accrued_interest

    @property
    def ytm(self):
        return self.__ytm

    @ytm.setter
    def ytm(self, ytm):
        if ytm is None:
            self.__ytm = ytm
        else:
            try:
                self.__ytm = float(ytm)
            except:
                raise ValueError('Bond class error: ' + ytm + ' is not'
                                 + ' a number')

    @property
    def priceper100(self):
        return self.__priceper100

    @priceper100.setter
    def priceper100(self, priceper100):
        if priceper100 is None:
            self.__priceper100 = priceper100
        else:
            try:
                self.__priceper100 = float(priceper100)
            except:
                raise ValueError('Bond class error: ' + priceper100
                                 + ' is not a number')

    @property
    def face_value(self):
        return self.__face_value

    @face_value.setter
    def face_value(self, face_value):
        if face_value is None:
            self.__face_value = face_value
        else:
            try:
                self.__face_value = float(face_value)
            except:
                raise ValueError('Bond class error: ' + face_value
                                 + ' is not a valid number for position')

    @property
    def proceed(self):
        return self.__proceed

    @property
    def risk_stats(self):
        return self.__risk_stats

    @risk_stats.setter
    def risk_stats(self, risk_stats):
        if risk_stats is None:
            self.__risk_stats = {}
        else:
            if isinstance(risk_stats, dict):
                self.__risk_stats = risk_stats
            else:
                raise ValueError('Bond class error: argument passed is not'
                                 + ' a dictionary.')

    @property
    def full_structures(self):
        return self.__full_structures

    @full_structures.setter
    def full_structures(self, full_structures):
        if full_structures is None:
            self.__full_structures = []
        else:
            if isinstance(full_structures, list):
                self.__full_structures = full_structures
            else:
                raise ValueError('Bond class error: argument passed is not'
                                 + ' a dictionary.')

    @property
    def active_structures(self):
        return self.__active_structures

    @active_structures.setter
    def active_structures(self, active_structures):
        if active_structures is None:
            self.__structures = []
        else:
            if isinstance(active_structures, list):
                self.__active_structures = active_structures
            else:
                raise ValueError('Bond class error: argument passed is not'
                                 + ' a Structures class.')

    def __get_bonddict(self):
        bonddict = {}
        bonddict['issue_date'] = self.issue_date
        bonddict['maturity'] = self.maturity
        bonddict['value_date'] = self.value_date
        bonddict['date_generation'] = self.date_gen
        bonddict['frequency'] = self.frequency
        bonddict['day_count'] = self.day_count
        bonddict['business_day'] = self.business_day
        bonddict['coupon'] = self.coupon
        bonddict['face_value'] = self.face_value
        return bonddict


    def construct_bond(self):
        bonddict = self.__get_bonddict()
        try:
            construct_fixbond(bonddict)
            # print(bonddict)

            self.full_structures = bonddict['full_structures']
            self.active_structures = bonddict['active_structures']
        except ValueError as error:
            print(str(error))


    def calculate(self, reconstruct=False):
        bonddict = self.__get_bonddict()
        bonddict['full_structures'] = self.full_structures
        if reconstruct is True:
            self.construct_bond()

        if self.__ytm is None:
            raise ValueError('ytm value is not assigned')

        try:
            self.recalc_act_structures()
            bonddict['active_structures'] = self.active_structures
            fixbond_price(self.value_date, bonddict, self.ytm)
        except ValueError as myerror:
            print(str(myerror))

        self.__accrued_interest = bonddict['accrued_interest']
        self.__risk_stats['pvbp01'] = bonddict['pvbp01']
        self.__risk_stats['modified_duration'] = bonddict['modified duration']
        self.__risk_stats['macaulay_duration'] = bonddict['macaulay duration']
        self.__risk_stats['duration'] = bonddict['duration']
        self.__risk_stats['convexity'] = bonddict['convexity']
        self.__active_structures = bonddict['active_structures']
        self.__priceper100 = bonddict['price per 100']
        self.__proceed = bonddict['proceed']
        self.last_coupon_date = bonddict['last_coupon_date']


    def recalc_act_structures(self):
        if self.value_date is None:
            raise Exception("Fixed Bond need value date for constructing" +
                            " active structures")

        # an error when issue date is none is not handle
        elif self.value_date < self.issue_date:
            acs = list(self.full_structures)

        else:
            fs = self.full_structures
            booldate = list(map(lambda x: x['date'] > self.value_date, fs))
            index = booldate.index(True)  # exclude paid coupon period
            acs = list(fs[index - 1:])

            acs[0]['dcf'] = 0
            acs[1]['dcf'] = day_cf(self.day_count, self.value_date,
                                   acs[1]['date'],
                                   bondmat_date=acs[-1]['date'],
                                   next_coupon_date=acs[1]['date'],
                                   Frequency=12 /
                                   frequencies[self.frequency])
            alen = len(acs)

            for i in range(alen):
                acs[i]['time'] = day_cf('Actual/365', self.value_date,
                                        acs[i]['date'],
                                        bondmat_date=acs[-1]['date'],
                                        next_coupon_date=acs[i]['date'])
            self.active_structures = acs


class FRN(Bond):

    def __init__(self):
        super().__init__()
        self.__risk_stats = {}
        self.__full_structures = {}
        self.average_period = 1
        self.fixing_basis = 'Same Day'


    @property
    def margin(self):
        return self.__margin

    @margin.setter
    def margin(self, margin):
        if margin is None:
            self.__margin = margin
        else:
            try:
                self.__margin = float(margin)
            except:
                self.__margin = None
                raise ValueError('FRN class error: ' + margin + ' is not'
                                 + ' a number')

    @property
    def spread(self):
        return self.__spread

    @spread.setter
    def spread(self, spread):
        if spread is None:
            self.__spread = spread
        else:
            try:
                self.__spread = float(spread)
            except:
                raise ValueError('FRN class error: ' + spread + ' is not'
                                 + ' a number')

    @property
    def fixing_basis(self):
        return self.__fixing_basis

    @fixing_basis.setter
    def fixing_basis(self, fixing_basis):
        if fixing_basis is None:
            self.__fixing_basis = 'Same Day'
        elif fixing_basis in start_basis:
            self.__fixing_basis = fixing_basis
        else:
            self.__fixing_basis = 'Same Day'

    @property
    def average_period(self):
        return self.__average_period

    @average_period.setter
    def average_period(self, average_period):
        if average_period is None:
            self.__average_period = average_period
        else:
            try:
                self.__average_period = int(average_period)
            except:
                raise ValueError('FRN class error: ' + average_period +
                                 ' is not a number')

    @property
    def current_coupon(self):
        return self.__current_coupon

    @current_coupon.setter
    def current_coupon(self, current_coupon):
        if current_coupon is None:
            self.__current_coupon = current_coupon
        else:
            try:
                self.__current_coupon = float(current_coupon)
            except:
                raise ValueError('FRN class error: ' + current_coupon +
                                 ' is not a number')

    @property
    def accrued_interest(self):
        return self.__accrued_interest

    @property
    def priceper100(self):
        return self.__priceper100

    @priceper100.setter
    def priceper100(self, priceper100):
        if priceper100 is None:
            self.__priceper100 = priceper100
        else:
            try:
                self.__priceper100 = float(priceper100)
            except:
                raise ValueError('FRN class error: ' + priceper100
                                 + ' is not a number')

    @property
    def face_value(self):
        return self.__face_value

    @face_value.setter
    def face_value(self, face_value):
        if face_value is None:
            self.__face_value = face_value
        else:
            try:
                self.__face_value = float(face_value)
            except:
                raise ValueError('Bond class error: ' + face_value
                                 + ' is not a valid number for face value')

    @property
    def proceed(self):
        return self.__proceed

    @property
    def risk_stats(self):
        return self.__risk_stats

    @risk_stats.setter
    def risk_stats(self, risk_stats):
        if risk_stats is None:
            self.__risk_stats = {}
        else:
            if isinstance(risk_stats, dict):
                self.__risk_stats = risk_stats
            else:
                raise ValueError('Bond class error: argument passed is not'
                                 + ' a dictionary.')

    @property
    def full_structures(self):
        return self.__full_structures

    @full_structures.setter
    def full_structures(self, full_structures):
        if full_structures is None:
            self.__full_structures = []
        else:
            if isinstance(full_structures, list):
                self.__full_structures = full_structures
            else:
                raise ValueError('Bond class error: argument passed is not'
                                 + ' a dictionary.')

    @property
    def active_structures(self):
        return self.__active_structures

    @active_structures.setter
    def active_structures(self, active_structures):
        if active_structures is None:
            self.__active_structures = []
        else:
            if isinstance(active_structures, list):
                self.__active_structures = active_structures
            else:
                raise ValueError('Bond class error: argument passed is not'
                                 + ' a Structures class.')

    def __get_bonddict(self):
        bonddict = {}
        bonddict['issue_date'] = self.issue_date
        bonddict['maturity'] = self.maturity
        bonddict['value_date'] = self.value_date
        bonddict['date_generation'] = self.date_gen
        bonddict['frequency'] = self.frequency
        bonddict['day_count'] = self.day_count
        bonddict['business_day'] = self.business_day
        bonddict['face_value'] = self.face_value
        bonddict['margin'] = self.margin
        bonddict['spread'] = self.spread
        bonddict['current_coupon'] = self.current_coupon
        bonddict['fixing_basis'] = self.fixing_basis
        bonddict['average_period'] = self.average_period
        return bonddict


    def construct_bond(self):
        bonddict = self.__get_bonddict()
        try:
            construct_frn(bonddict)
            self.__full_structures = bonddict['full_structures']
            self.__active_structures = bonddict['active_structures']
        except ValueError as error:
            print(str(error))


    def recalc_act_structures(self):
        if self.value_date is None:
            raise Exception("FRN need value date for constructing" +
                            " active structures")

        elif self.value_date < self.issue_date:
            acs = list(self.full_structures)

        else:
            fs = self.full_structures
            booldate = list(map(lambda x: x['date'] > self.value_date, fs))
            index = booldate.index(True)  # exclude paid coupon period
            acs = list(fs[index - 1:])

            acs[0]['y_dcf'] = 0
            acs[0]['fixed'] = None
            acs[1]['y_dcf'] = day_cf(self.day_count, self.value_date,
                                   acs[1]['date'],
                                   bondmat_date=acs[-1]['date'],
                                   next_coupon_date=acs[1]['date'],
                                   Frequency=12 /
                                   frequencies[self.frequency])
            alen = len(acs)

            for i in range(alen):
                row = acs[i]
                row['time'] = day_cf('Actual/365', self.value_date,
                                        row['date'],
                                        bondmat_date=acs[-1]['date'],
                                        next_coupon_date=row['date'])
                if row['fixing_date'] is not None:
                    if ((row['fixing_date'] <= self.value_date) and (
                            self.value_date < row['date'])):
                        row['fixed'] = True
                        row['coupon'] = self.current_coupon
                    else:
                        row['fixed'] = False

            self.active_structures = acs


    def calculate(self, disc_curve, reconstruct=False):
        """


        Parameters
        ----------
        disc_curve : DFactor
            DFactor class containg information on the discount factors.
        reconstruct : boolean, optional
            Constructing the bonds coupon structure if set to true.
            The default is False.

        Returns
        -------
        None

        """
        bonddict = self.__get_bonddict()
        if reconstruct is True:
            self.construct_bond()
        try:
            self.recalc_act_structures()
            bonddict['active_structures'] = self.active_structures
            frn_price(bonddict, disc_curve)
        except ValueError as myerror:
            print(str(myerror))

        self.__accrued_interest = bonddict['accrued_interest']
        self.__risk_stats['pvbp01'] = bonddict['pvbp01']
        #self.__risk_stats['modified_duration'] = bonddict['modified duration']
        #self.__risk_stats['macaulay_duration'] = bonddict['macaulay duration']
        #self.__risk_stats['duration'] = bonddict['duration']
        #self.__risk_stats['convexity'] = bonddict['convexity']
        #self.__active_structures = bonddict['active_structures']
        self.__priceper100 = bonddict['price per 100']
        self.__proceed = bonddict['proceed']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 30 13:36:21 2020

@author: RMS671214
"""

from interestrate import discount_curve as dcurve

# %%
rates = []
rate = {"value_date": "2020-10-30", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '12M': 2.75, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}
rates.append(rate)
rate = {"value_date": "2020-10-30", "st_busday": "Modified Following",
        "st_ratebasis": "Simple", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 3.30, '1W': 3.35, '1M': 3.45, '3M': 3.55,
                  '6M': 3.65, '12M': 3.75, '1Y': 3.70, '2Y': 3.80, '3Y': 3.90,
                  '5Y': 4.00, '10Y': 4.10, '30Y': 4.25}}
rates.append(rate)
print(rates)

dfs = dcurve.discount_factor_gen(rate, return_type="times")
print(dfs)

# %%

flat = dcurve.flat_curve("2020-10-30", "2020-12-30", 5.30, rate_basis="Money Market",
                         day_count="Actual/365", bus_day="No Adjustment",
                         tenors=None, return_type="days")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  5 14:42:51 2020

@author: RMS671214
"""

from faspy.basel.liquidity import hqla
# Top Level
hqlass = hqla.Level("HQLA", None)
hqlass.child["Level1"] = hqla.Level("Level1", None)
# Level 1
lvl1 = hqlass.child["Level1"]
lvl1.child['Coins'] = hqla.Level("Coins and Bank Notes", 1.00)
lvl1.child['Coins'].amount = 1_000_000_000

lvl1.child['BNM_Reserves'] = hqla.Level("BNM Reserves", 1.00)
lvl1.child['BNM_Reserves'].amount = 2_000_000_000

lvl1.child['Securities'] = hqla.Level("Marketable Securities", 1.00)
lvl1.child['Securities'].amount = 3_000_000_000

lvl1.child['Non_MYR_Assets'] = hqla.Level("Non MYR Assets", 1.00)
lvl1.child['Non_MYR_Assets'].amount = 4_000_000_000

lvl1.child['MYR_Assets'] = hqla.Level("MYR Assets", 1.00)
lvl1.child['MYR_Assets'].amount = 5_000_000_000

# Level2
hqlass.child["Level2"] = hqla.Level("Level2", None)
level2 = hqlass.child["Level2"]
level2.child["Level2A"] = hqla.Level("Level2A", None)
level2A = level2.child["Level2A"]
level2A.child["Cagamas"] = hqla.Level("Cagamas", 0.85)
level2A.child["Cagamas"].amount = 4_000_000_000

level2A.child["PDS"] = hqla.Level("PDS", 0.85)
level2A.child["PDS"].amount = 2_000_000_000

level2.child["Level2B"] = hqla.Level("Level2B", None)
level2B = level2.child["Level2B"]
level2B.child["RMBS"] = hqla.Level("RMBS", 0.75)
level2B.child["RMBS"].amount = 1_000_000_000

level2B.child["PDS"] = hqla.Level("PDS", 0.50)
level2B.child["PDS"].amount = 2_000_000_000

level2B.child["Equity"] = hqla.Level("Equity", 0.50)
level2B.child["Equity"].amount = 3_000_000_000

# Cash Outflow
oflow = hqla.Level("Outflow", None)
oflow.child["retail_deposit"] = hqla.Level("A. Retail Deposits", None)
rdepo = oflow.child["retail_deposit"]
rdepo.child["deposits"] = hqla.Level("Demand deposits and term seposits " +
                                         "(less than 30 days maturity", None)
depo = rdepo.child["deposits"]
depo.child["stable_depo"] = hqla.Level("Stable deposits (deposit insurance " +
                                           "scheme meeting additional criteria)", 0.03)
depo.child["stable_depo"].amount = 1_000_000_000

depo.child["stable_depo2"] = hqla.Level("Stable deposits ", 0.05)
depo.child["stable_depo2"].amount = 4_000_000_000

depo.child["stable_depo3"] = hqla.Level("Less stable retail deposits ", 0.1)
depo.child["stable_depo3"].amount = 250_000_000

rdepo.child["term_deposit"] = hqla.Level("Term deposits with residual maturity " +
                                             "greater than 30 days)", 0.00)

oflow.child["unsecured_wholesale"] = hqla.Level("B. Unsecured wholesale funding", None)
unsecured = oflow.child["unsecured_wholesale"]
unsecured.child["deposits"] = hqla.Level("Demand and term deposits(less than " +
                                             "30 days maturity) provided by " +
                                             "small business customers", None)
udepo = unsecured.child["deposits"]
udepo.child["stable_deposits"] = hqla.Level("Stable Deposits", 0.05)
udepo.child["stable_deposits"].amount = 2_000_000_000

udepo.child["less_stable_deposits"] = hqla.Level("Less Stable Deposits", 0.1)
udepo.child["less_stable_deposits"].amount = 500_000_000
desc = "".join(["Operational deposits generated by clearing, custody and cash ",
                "management activities"])
unsecured.child["operational_deposit"] = hqla.Level(desc, 0.25)
unsecured.child["operational_deposit"].amount = 75_000_000
odepo = unsecured.child["operational_deposit"]
odepo.child["deposit_insurance"] = hqla.Level(desc, 0.05)
odepo.child["deposit_insurance"].amount = 150_000_000

desc= "Cooperative banks in an institutional network (qualifying deposits with the centralised institution)"
unsecured.child["cooperative"] = hqla.Level(desc, 0.25)
unsecured.child["cooperative"].amount = 2_000_000_000

desc = "Non-financial corporates, sovereigns, central banks, multilateral development banks, and PSEs"
unsecured.child["nonfinancial"] = hqla.Level(desc, 0.4)
unsecured.child["nonfinancial"].amount = 750_000_000

desc = "If the entire amount fully covered by deposit insurance scheme"
nonfin = unsecured.child["nonfinancial"]
nonfin.child["deposit_insurance"] = hqla.Level(desc, 0.25)
nonfin.child["deposit_insurance"].amount = 0

desc = "Other legal entity customers"
unsecured.child["others"] = hqla.Level(desc, 1.00)
unsecured.child["others"].amount = 100_000_000

oflow.child["secured"] = hqla.Level("C. Secured Funding", None)
secured = oflow.child["secured"]
desc = "Secured funding transactions with a central bank counterparty or backed by Level 1 assets with any counterparty."
secured.child["central_banks"] = hqla.Level(desc, 0.00)
secured.child["central_banks"].amount = 0

desc = "Secured funding transactions backed by Level 2A assets, with any counterparty"
secured.child["level2A_assets"] = hqla.Level(desc, 0.15)
secured.child["level2A_assets"].amount = 0

desc = "• Secured funding transactions backed by non-Level 1 or non-Level 2A assets, with domestic sovereigns, multilateral development banks, or domestic PSEs as a counterparty"
secured.child["non_level2A_assets"] = hqla.Level(desc, 0.25)
secured.child["non_level2A_assets"].amount = 0

desc = "Backed by RMBS eligible for inclusion in Level 2B"
secured.child["level2B_RMBS"] = hqla.Level(desc, 0.25)
secured.child["level2B_RMBS"].amount = 0

desc = "Backed by other Level 2B assets"
secured.child["level2B_others"] = hqla.Level(desc, 0.50)
secured.child["level2B_others"].amount = 0

desc = "All other secured funding transactions"
secured.child["others"] = hqla.Level(desc, 1.00)
secured.child["others"].amount = 0

oflow.child["additional"] = hqla.Level("D. Additional requirements:", None)
add = oflow.child["additional"]
desc = "Liquidity needs (eg collateral calls) related to financing transactions, derivatives and other contracts"
add.child["liquidity"] = hqla.Level(desc, None)

desc = "Market valuation changes on derivatives transactions (largest absolute net 30-day collateral flows realised during the preceding 24 months)"
add.child["derivative_changes"] = hqla.Level(desc, None)

desc = "Valuation changes on non-Level 1 posted collateral securing derivatives"
add.child["non_level1_changes"] = hqla.Level(desc, 0.2)

desc = "Excess collateral held by a bank related to derivative transactions that could contractually be called at any time by its counterparty"
add.child["excess_coll"] = hqla.Level(desc, 1.00)

desc = "Liquidity needs related to collateral contractually due from the reporting bank on derivatives transactions"
add.child["coll_liq"] = hqla.Level(desc, 1.00)

desc = "Increased liquidity needs related to derivative transactions that allow collateral substitution to non-HQLA assets"
add.child["inc_liquidity"] = hqla.Level(desc, 1.00)

desc = "ABCP, SIVs, conduits, SPVs, etc:"
add.child["conduits"] = hqla.Level(desc, None)
cond = add.child["conduits"]
desc = "Liabilities from maturing ABCP, SIVs, SPVs, etc (applied to maturing amounts and returnable assets)"
cond.child["liabilities"] = hqla.Level(desc, 1.00)

desc = "• Asset Backed Securities (including covered bonds) applied to maturing amounts."
cond.child["assets"] = hqla.Level(desc, 1.00)

desc = "Currently undrawn committed credit and liquidity facilities provided to:"
add.child["uncommitted"] = hqla.Level(desc, 1.00)
unc = add.child["uncommitted"]

desc = "retail and small business clients"
unc.child["retail"] = hqla.Level(desc, 0.05)

desc = "• non-financial corporates, sovereigns and central banks, multilateral development banks, and PSEs"
unc.child["non_fi"] = hqla.Level(desc, None)
nonfi = unc.child["non_fi"]
nonfi.child["credit"] = hqla.Level("Credit", 0.1)
nonfi.child["liquidity"] = hqla.Level("Liquidity", 0.3)

desc = "• banks subject to prudential supervision"
unc.child["prudent"] = hqla.Level(desc, 0.4)
prudent = unc.child["prudent"]

desc = " other financial institutions (include securities firms, insurance companies)"
unc.child["other_fi"] = hqla.Level(desc, None)
ofi = unc.child["other_fi"]
ofi.child["credit"] = hqla.Level("Credit", 0.4)
ofi.child["liquidity"] = hqla.Level("Liquidity", 1.00)

desc = "• other legal entity customers, credit and liquidity facilities"
unc.child["other_le"] = hqla.Level(desc, 1.00)

desc = "Other contingent funding liabilities (such as guarantees, letters of credit, revocable credit and liquidity facilities, etc)"
add.child["others"] = hqla.Level(desc, None)
oth = add.child["others"]

oth.child["trade_fin"] = hqla.Level("Trade Finance", 0.05)
desc = "Customer short positions covered by other customers’ collateral"
oth.child["other_coll"] = hqla.Level(desc, 0.50)

desc = "Any additional contractual outflows"
add.child["add_outflow"] = hqla.Level(desc, 1.00)

desc = "Net derivative cash outflows"
add.child["der_outflow"] = hqla.Level(desc, 1.00)

desc = "Any other contractual cash outflows"
add.child["other_cash_outflow"] = hqla.Level(desc, 1.00)

# Cash Inflow
iflow = hqla.Level("Outflow", None)
iflow.child["level1"] = hqla.Level("Level 1 assets", 0.00)
iflow.child["level2A"] = hqla.Level("Level 2A assets", 0.15)
iflow.child["level2A"].amount = 200_000_000
iflow.child["level2B"] = hqla.Level("Level 2B assets", None)
lvl2b = iflow.child["level2B"]
lvl2b.child["RMBS"] = hqla.Level("Eligible RMBS", 0.25)
lvl2b.child["Others"] = hqla.Level("Other assets", 0.50)
lvl2b.child["Others"].amount = 50_000_000

desc = "Margin lending backed by all other collateral"
iflow.child["margin"] = hqla.Level(desc, 0.50)
iflow.child["margin"].amount = 50_000_000

desc = "All other assets"
iflow.child["other_assets"] = hqla.Level(desc, 1.00)
iflow.child["other_assets"].amount = 25_000_000

desc = "Credit or liquidity facilities provided to the reporting bank"
iflow.child["cr_liq"] = hqla.Level(desc, 0.00)
iflow.child["cr_liq"].amount = 10_000_000_000

desc = "Operational deposits held at other financial institutions (include deposits held at centralised institution of network of co-operative banks)"
iflow.child["op_depo"] = hqla.Level(desc, 0.00)
iflow.child["op_depo"].amount = 1_000_000_000

desc = "Other inflows by counterparty:  "
iflow.child["other_inflows"] = hqla.Level(desc, None)
oiflow = iflow.child["other_inflows"]
desc = "Amounts to be received from retail counterparties "
oiflow.child["retail"] = hqla.Level(desc, 0.5)
oiflow.child["retail"].amount = 50_000_000

desc = "Amounts to be received from non-financial wholesale counterparties, from transactions other than those listed in above inflow categories"
oiflow.child["wholesale"] = hqla.Level(desc, 0.5)
oiflow.child["wholesale"].amount = 70_000_000

desc = "Amounts to be received from financial institutions and central banks, from transactions other than those listed in above inflow categories"
oiflow.child["fis"] = hqla.Level(desc, 1.00)
oiflow.child["fis"].amount = 80_000_000

desc = "Net derivative cash inflows"
iflow.child["def_flows"] = hqla.Level(desc, 1.00)
iflow.child["def_flows"].amount = 25_000_000

desc = "Other contractual cash inflows"
iflow.child["others"] = hqla.Level(desc, None)

print(iflow.value())
print(iflow.get_array())

# %%
import pandas as pd
pdhqla = pd.DataFrame(hqlass.get_array())







import numba
import numpy as np



@numba.njit('float64(float64,float64,float64,float64)')
def bond_price(tenor, yld, cpn, fre):

    total_cpn = np.int32(tenor * fre)
    cpn_values = 0.00
    for period in range(1,total_cpn + 1, 1):
        cpn_value = (cpn / fre) / (1 + (yld * 0.01) / fre) ** (period)
        cpn_values += cpn_value


    pv_fv = 100 / (1 + (yld  * 0.01) / fre ) ** (tenor * fre)
    price = cpn_values + pv_fv
    #if yld == 5 :
    #    print(tenor, yld, cpn, fre, price)
    return price


@numba.njit('float64(float64,float64,float64,float64)')
def bond_duration(tenor, yld, cpn, fre):
    inc = 0.0001
    price0 = bond_price(tenor, yld, cpn, fre)
    price1 = bond_price(tenor, yld + inc, cpn, fre)
    duration = (price1 - price0)/inc
    return duration


@numba.njit('float64(float64,float64,float64,float64)')
def bond_convexity(tenor, yld, cpn, fre):
    inc = 0.0001
    duration0 = bond_duration(tenor, yld, cpn, fre)
    duration1 = bond_duration(tenor, yld + inc, cpn, fre)
    convexity = (duration1 - duration0)/inc
    return convexity



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 30 13:36:21 2020

@author: RMS671214
"""

from faspy.interestrate import discount_curve as dcurve
from faspy.interestrate import rmp_curves as rcurve

# %%

rate = {"value_date": "2020-10-30", "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '12M': 2.75, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}


dfs = dcurve.discount_factor_gen(rate, return_type="times")


# %%

x_axis = [x["times"] for x in dfs]
y_axis = [x["df"] for x in dfs]

start = 0.5
end = 0.75
# use this to calculate a fwd discount factor
fwd1 = rcurve.calc_fwd_df(start, end, x_axis, y_axis)

ifunc = rcurve.interpolation(x_axis, y_axis, start, is_function=True)
print(type(ifunc))
# use this to calculate a long list of forward discount factor
fwd2 = rcurve.calc_fwd_df(start, end, ifunc=ifunc)

# e.g of using list
fwd_period = [[0.1, 0.2], [0.2, 0.3], [0.4, 0.7], [0.7, 0.9]]
fwd_dfs = [rcurve.calc_fwd_df(x[0], x[1], ifunc=ifunc) for x in fwd_period]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 21:02:58 2020

@author: RMS671214
"""

from faspy.interestrate.fixincome import date_structures, calc_customfix_structures, \
    value_customfix_structures
import numpy as np
from faspy.interestrate import rmp_dates as rd
from faspy.interestrate import discount_curve as dcurve
# %%

mybond = {}
mybond['issue_date'] = np.datetime64('2018-10-22')
mybond['value_date'] = np.datetime64('2021-10-22')
mybond['maturity'] = np.datetime64('2028-10-22')
mybond['day_count'] = 'Actual/365 Fixed'
mybond['frequency'] = 'Semi-Annual'
mybond['business_day'] = 'No Adjustment'
mybond['date_generation'] = rd.date_gen_method[1]
mybond['face_value'] = 1000000
mybond['coupon'] = 10
mybond['ytm'] = None
mybond['type'] = 'Fixed Rate Bond'

structures = list(date_structures(mybond))
print("DATES")
print("=====")
print(structures)

try:
    import pandas as pd
    pd1 = pd.DataFrame(structures)
except:
    pass


# %%
# Create Sample structure

lenstruc = len(structures)
for i in range(lenstruc):
    structure = structures[i]
    if i == 0:
        structure["face_value"] = 10_000_000
        structure["fv_flow"] = 1_000_000
        structure["coupon"] = 5.00
    else:
        prevstruc = structures[i-1]
        structure["face_value"] = prevstruc["face_value"] + prevstruc["fv_flow"]
        structure["fv_flow"] = 1_000_000
        structure["coupon"] = prevstruc["coupon"] + 0.25

structures[-1]["fv_flow"] = structures[-1]["face_value"]
new_structures = calc_customfix_structures(structures,
                                           mybond["day_count"],
                                           mybond["frequency"],
                                           mybond["business_day"])
print("CUSTOM STRUCTURE")
print("================")
print(new_structures)
try:
    import pandas as pd
    pd2 = pd.DataFrame(new_structures)
except:
    pass

# %%
# valuing the structures using flat curve

flatcurve = dcurve.flat_curve(mybond['value_date'], mybond['value_date'] + 50,
                              3.5)
print(flatcurve)

newstruct = value_customfix_structures(mybond["value_date"], new_structures,
                                       mybond["day_count"],
                                       mybond["frequency"], flatcurve)
print("VALUING CUSTOM STRUCTURE")
print("========================")
print(newstruct)
try:
    import pandas as pd
    pd3 = pd.DataFrame(newstruct)
except:
    pass


# %%

coupon_dates = list(date_structures(mybond))
print(coupon_dates)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 16 14:11:12 2020

@author: RMS671214
"""
import math
import numba
from time import time

@numba.njit('float64(float64)')
def rmpcdf(x):
    return 0.5 * (1 + math.erf(x/(2**0.5)))

@numba.njit('float64(float64)')
def rmpppf(x):
    cd = 0.75
    value0 = rmpcdf(cd)
    value1 = rmpcdf(cd + 0.000000001)
    der = (value1 - value0) / 0.000000001
    counter = 0
    while counter <500:
        cd = cd - (value0 - x)/der
        value0 = rmpcdf(cd)
        value1 = rmpcdf(cd + 0.000000001)
        if abs(value0-x) < 0.00000000000000000000001:
            counter = 500

        der = (value1 - value0) / 0.000000001
        counter += 1

    return cd



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 29 09:39:29 2020

@author: RMS671214
"""

from .rmp_dates import  day_count_factor as day_cf
from .rmp_dates import tenor_to_maturity as ttm
from .rmp_curves import generate_st_df as gen_stdf
from numpy import datetime64 as dt64, busdaycalendar as cal
from numpy import busday_offset as bus_off
from .rmp_curves import generate_fulldf as gen_fdf, \
    interpolation
from numba import njit
from .conventions import *

class Data:
    """
    A class to store single data by means of a single level/depth dictionary

    Attributes:
    ===========
    data: dict
        Single level dictionary. There can be various number of keys.
        Values can be any object required subject to the limitation that
        the class is meant for single level dictionary

    Methods:
    ========
    single_data:
        return the value for the key provided.
    """

    def __init__(self, data=None):
        # data should be a dictionary like {'O/N': True, 'T/N':False}
        # the dict may hold value for rate, boolean etc depending on the intended
        # use of the class
        self.data = data

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, newdata):
        if isinstance(newdata, dict):
            self.__data = newdata
        else:
            self.__data = {}


    def get(self, data_name):
        """
        Parameters
        ----------
        data_name : str
            key for the data attributes

        Returns
        -------
        obj
            Returns the object held by the key in the atributes. If the key
            is not found, it will return None.

        """
        return self.__data.get(data_name)




class HistoricalData:
    """

    A collection class for Data. It hold multiple values/records of Data object
    Attributes:
    ===========
    historical: dict
        A dictionary containing Data object

    Methods:
        addnewdata:
            method to add new data to the class

        view_data_as_dict:
            return multilevel dictionary of historical attribute

    """
    def __init__(self):
        # historical is a dictionary of Data
        self.historical = {}

    @property
    def historical(self):

        return self.__historical

    @historical.setter
    def historical(self,historical):

        if historical is None or isinstance(historical,dict):
            self.__historical = historical
        else:
            raise Exception ('historical data is not in a dictionary')

    def addnewdata(self,newdata):

        #historical is a dictionary of Data
        if isinstance(newdata,dict):
            for key in newdata:
                self.__historical[key] = newdata[key]
        else:
            raise Exception('HistoricalRates error: data given for addnewdata is not a dictionary')


    def view_data_as_dict(self):
        """
        converts historical attribute to full-fledge dictionary
        Returns
        -------
        result : dict
            The whole content of historical attributes as dict object

        """
        result = {}
        for key in self.__historical:
            tenor_data = self.__historical[key]
            if isinstance(tenor_data, Data):
                result[key] = tenor_data.data
            else:
                result[key] = self.__historical[key]

        return result


# %%

class STRate():
    _basis = ('Simple', 'Discount Rate')

    def __init__(self, start_date, maturity, rate,
                 day_count='Actual/365 Fixed', rate_basis='Money Market',
                 business_day='No Adjustment'):
        self.start_date = start_date
        self.maturity = maturity
        self.day_count = day_count
        self.rate_basis = rate_basis
        self.business_day = business_day
        self.rate = rate


    def discount_factor(self):
        if self.day_count not in day_counts:
            return None
        elif self.rate_basis not in self._basis:
            return None
        elif self.business_day not in business_days:
            return None

        d_count = dc_fac(self.day_count, self.start_date, self.maturity)
        result = None
        if self.rate_basis == self._basis[0]:
            result = 1/(1 + self.rate * d_count * 0.01)
        elif self.rate_basis == self._basis[1]:
            result = 1 - self.rate * d_count * 0.01


        return result

# %%
class STRt(Data):
    __basis = ('Money Market', 'Discount Rate')
    __termsorder = {'O/N': 1, 'T/N': 2, 'S/N': 2.5, '1W': 3, '2W': 4, '3W': 5, '1M': 6,
                    '2M': 7, '3M': 8, '4M': 9, '5M': 10, '6M': 11, '9M': 15,
                    '12M': 20}
    __termsrates = {'O/N': None, 'T/N': None, 'S/N': None, '1W': None, '2W': None,
                    '3W': None, '1M': None, '2M': None, '3M': None,
                    '4M': None, '5M': None, '6M': None, '9M': None,
                    '12M': None}
    __spotbasis = ('Same Day', 'Tom', 'Spot')

    def __init__(self):
        self.value_date = None
        self.rate_basis = 'Money Market'
        self.day_count = None
        self.business_day = None
        self.data = {}
        self.__ratedict = dict(self.__termsrates)

    @property
    def current_date(self):
        return self.__current_date

    @current_date.setter
    def current_date(self, current_date):
        if current_date is not None:
            try:
                self.__current_date = dt64(current_date)
            except:
                raise ValueError('current date is not a date')
        else:
            self.__current_date = current_date

    @property
    def value_date(self):
        return self.__value_date

    @value_date.setter
    def value_date(self, value_date):
        if value_date is not None:
            try:
                self.__value_date = dt64(value_date)
            except:
                raise ValueError('start_date is not a date')
        else:
            self.__value_date = value_date

    @property
    def rate_basis(self):
        return self.__rate_basis

    @rate_basis.setter
    def rate_basis(self, rate_basis):

        if rate_basis is not None:
            if rate_basis in self.__basis:
                self.__rate_basis = rate_basis
            else:
                raise ValueError('rate_basis is not in the list of acceptable values')
        else:
            self.__rate_basis = 'Money Market'

    @property
    def start_basis(self):
        return self.__start_basis

    @start_basis.setter
    def start_basis(self, start_basis):

        if start_basis is not None:
            if start_basis in self.__spotbasis:
                self.__start_basis = start_basis
            else:
                raise ValueError('start_basis is not in the list of acceptable values')
        else:
            self.__start_basis = 'Same Day'

    @property
    def day_count(self):
        return self.__day_count

    @day_count.setter
    def day_count(self, day_count):
        if day_count is not None:
            if day_count in day_counts:
                self.__day_count = day_count
            else:
                raise ValueError('day_count is not in the list of acceptable values')
        else:
            self.__day_count = 'Actual/365'

    @property
    def business_day(self):
        return self.__business_day

    @business_day.setter
    def business_day(self, business_day):
        if business_day is not None:
            if business_day in business_days:
                self.__business_day = business_day
            else:
                raise ValueError('business_day is not in the list of acceptable values')
        else:
            self._business_day = 'Modified Following'

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, newdata):
        if isinstance(newdata, dict):
            tenors = []
            therates = []
            for k in newdata:
                if newdata.get(k) is not None:
                    self.__ratedict[k] = newdata[k]
                    tenors.append(k)
                    therates.append(newdata[k])
            self.__data = {'tenors': tenors, 'rates': therates}

        elif isinstance(newdata, Data):
            self.__data = newdata.data
        else:
            self.__data = {}

    def calcdf(self, holidays=[]):
        setting = {'cdate': self.__current_date,
                   'start_basis': self.start_basis,
                   'day_count': self.__day_count,
                   'bus_day': self.__business_day,
                   'rate_basis': self.__rate_basis}
        tenors =[]
        therates =[]
        for k in self.__ratedict:
            tenors.append(k)
            therates.append(self.__ratedict[k])
        result = _calc_shorttenor(tenors, therates,
                                  setting, holidays=holidays)
        return result


# %%
class STRates(Data):
    __basis = ('Simple', 'Discount Rate', 'Continuous')
    __termsrates = {'1W': None, '2W': None,
                    '3W': None, '1M': None, '2M': None, '3M': None,
                    '4M': None, '5M': None, '6M': None, '9M': None,
                    '12M': None}

    def __init__(self):
        self.start_date = None
        self.rate_basis = 'Simple'
        self.day_count = None
        self.business_day = None
        self.data = {}
        self.rates = dict(self.__termsrates)


    @property
    def start_date(self):
        return self.__start_date

    @start_date.setter
    def start_date(self, start_date):
        if start_date is not None:
            try:
                self.__start_date = dt64(start_date)
            except:
                raise ValueError('start_date is not a date')
        else:
            self.__start_date = start_date

    @property
    def rate_basis(self):
        return self.__rate_basis

    @rate_basis.setter
    def rate_basis(self, rate_basis):

        if rate_basis is not None:
            if rate_basis in self.__basis:
                self.__rate_basis = rate_basis
            else:
                raise ValueError('rate_basis is not in the list of acceptable values')
        else:
            self.__rate_basis = 'Simple'

    @property
    def day_count(self):
        return self.__day_count

    @day_count.setter
    def day_count(self, day_count):
        if day_count is not None:
            if day_count in day_counts:
                self.__day_count = day_count
            else:
                raise ValueError('day_count is not in the list of acceptable values')
        else:
            self.__day_count = 'Actual/365'

    @property
    def business_day(self):
        return self.__business_day

    @business_day.setter
    def business_day(self, business_day):
        if business_day is not None:
            if business_day in business_days:
                self.__business_day = business_day
            else:
                raise ValueError('business_day is not in the list of'
                                 + ' acceptable values')
        else:
            self._business_day = 'Modified Following'

    @property
    def rates(self):
        return self.__rates

    @rates.setter
    def rates(self, newdata):
        self.__rates = dict(self.__termsrates)
        if isinstance(newdata, dict):
            tenors = []
            therates = []
            for k in self.__termsrates:
                if newdata.get(k) is not None:
                    try:
                        float(newdata[k])
                        tenors.append(k)
                        therates.append(newdata[k])
                        self.__rates[k] = newdata[k]
                    except:
                        pass

        elif isinstance(newdata, Data):
            self.__rates = newdata.data

    def getdata(self):
        strates = {}
        for k in self.__rates:
            if self.__rates.get(k) is not None:
                strates[k] = self.__rates[k]

        return strates


    def calc_df(self):
        strates = self.getdata()
        result = gen_stdf(self.start_date, strates, self.day_count,
                          self.business_day, rate_basis=self.rate_basis)
        return result



# %%
class LTRates(Data):

    __termsrates = {'1Y': None, '2Y': None,
                    '3Y': None, '4Y': None, '5Y': None, '6Y': None,
                    '7Y': None, '10Y': None, '15Y': None, '20Y': None,
                    '30Y': None}

    def __init__(self):
        self.start_date = None
        self.frequency = 'Semi-Annual'
        self.day_count = None
        self.business_day = None
        self.__rates = dict(self.__termsrates)


    @property
    def start_date(self):
        return self.__start_date

    @start_date.setter
    def start_date(self,start_date):
        if start_date is not None:
            try:
                self.__start_date = dt64(start_date)
            except:
                raise ValueError('start_date is not a date')
        else:
            self.__start_date = start_date


    @property
    def frequency(self):
        return self.__frequency

    @frequency.setter
    def frequency(self,frequency):
        if frequency is not None:
            if frequency in frequencies:
                self.__frequency = frequency
            else:
                raise ValueError('frequency is not in the list of acceptable values')
        else:
            self.__frequency = 'Semi-Annual'


    @property
    def day_count(self):
        return self.__day_count

    @day_count.setter
    def day_count(self,day_count):
        if day_count is not None:
            if day_count in day_counts:
                self.__day_count = day_count
            else:
                raise ValueError('day_count is not in the list of acceptable values')
        else:
            self.__day_count = 'Actual/365'



    @property
    def business_day(self):
        return self.__business_day

    @business_day.setter
    def business_day(self,business_day):
        if business_day != None:
            if business_day in business_days:
                self.__business_day = business_day
            else:
                raise ValueError('business_day is not in the list of acceptable values')
        else:
            self._business_day = 'Modified Following'


    @property
    def rates(self):
        return self.__rates

    @rates.setter
    def rates(self, newdata):
        self.__rates = dict(self.__termsrates)
        if isinstance(newdata, dict):
            tenors = []
            therates = []
            for k in self.__termsrates:
                if newdata.get(k) is not None:
                    try:
                        float(newdata[k])
                        tenors.append(k)
                        therates.append(newdata[k])
                        self.__rates[k] = newdata[k]
                    except:
                        pass


        elif isinstance(newdata, Data):
            self.__rates = newdata.data


    def getdata(self):
        strates = {}
        for k in self.__rates:
            if self.__rates.get(k) is not None:
                strates[k] = self.__rates[k]

        return strates


#%%
class Rates:
    def __init__(self):
        self.strates = STRates()
        self.ltrates = LTRates()
        self.__value_date = None
        self.date_gen_method = date_gen_method[0]
        self.holidays = []
        self.__df = DFactor()
        self.__par = {}

    @property
    def value_date(self):
        return self.__value_date

    @value_date.setter
    def value_date(self, value_date):
        if value_date is not None:
            try:
                self.__value_date = dt64(value_date, "D")
            except:
                raise ValueError('value_date is not a date')
        else:
            self.__value_date = value_date

    @property
    def df(self):
        return self.__df

    @property
    def par(self):
        return self.__par


    def calcdf(self):
        st = self.strates
        lt = self.ltrates

        st_curve = st.getdata()
        lt_curve = lt.getdata()

        all_df = gen_fdf(self.value_date, st_curve, st.day_count,
                          st.business_day, st.rate_basis, lt_curve,
                          lt.day_count, lt.business_day,
                          frequencies[lt.frequency],
                          method=self.date_gen_method,
                          holidays=self.holidays)

        self.df.data = list(all_df)
        # print(list(all_df))
        #self.__par = all_df['par_rates']


    def interpolate(self, x_value, model='chip', method='slinear'):
        disfac = self.df.data
        x_axis = [x['time'] for x in disfac]
        #x_axis = self.par['times']
        y_axis = [y['df'] for y in disfac]
        #y_axis =self.par['rates']
        return interpolation(x_axis, y_axis, x_value, model=model,
                             method=method)


#%%
class DFactor(Data):
    def __init__(self):
        self.times = None
        self.day_counts = None
        self.rates = None
        self.dfs = None
        self.dates = None
        self.tenors = None
        self.data = []

    @property
    def day_counts(self):
        return self.__day_counts

    @day_counts.setter
    def day_counts(self,day_counts):
        self.__day_counts = day_counts

    @property
    def times(self):
        return self.__times

    @times.setter
    def times(self,times):
        self.__times = times


    @property
    def dfs(self):
        return self.__dfs

    @dfs.setter
    def dfs(self,dfs):
        self.__dfs = dfs

    @property
    def rates(self):
        return self.__rates

    @rates.setter
    def rates(self, rates):
        self.__rates = rates



    @property
    def dates(self):
        return self.__dates

    @dates.setter
    def dates(self, dates):
        self.__dates = dates


    @property
    def tenors(self):
        return self.__tenors

    @tenors.setter
    def tenors(self, tenors):
        self.__tenors = tenors



    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, newdata):
        if isinstance(newdata, list):
            self.__data = newdata
        elif isinstance(newdata, Data):
            self.__data = newdata.data
        else:
            self.__data = []

        #self._assign_attrs()



    def _assign_attrs(self):

        self.__times = self.__data.get('times')
        self.__dates = self.__data.get('dates')
        self.__rates = self.__data.get('rates')
        self.__dfs = self.__data.get('dfs')
        self.__tenors = self.__data.get('tenors')
        self.__day_counts = self.__data.get('dcfs')


    def interpolate(self, x_value, x_axis='times', model='chip',
                    method='slinear',is_function=False):
        if x_axis == 'times':
            myx = list(map(lambda x: x['time'], self.__data))
        elif x_axis == 'day_count':
            myx = list(map(lambda x: x['dcf'], self.__data))

        y_axis = list(map(lambda x: x['df'], self.__data))
        return interpolation(myx, y_axis, x_value, model=model, method=method, is_function=is_function)


class SwapPoint():
    __termsorder = {'O/N': 1, 'T/N': 2, 'S/N': 2.5, '1W': 3, '2W': 4, '3W': 5, '1M': 6,
                    '2M': 7, '3M': 8, '4M': 9, '5M': 10, '6M': 11, '9M': 15,
                    '12M': 20}
    __spotbasis = ('Same Day', 'Tom', 'Spot')
    __termsrates = {'O/N': None, 'T/N': None, 'S/N': None, '1W': None, '2W': None,
                    '3W': None, '1M': None, '2M': None, '3M': None,
                    '4M': None, '5M': None, '6M': None, '9M': None,
                    '12M': None}
    def __init__(self):
        self.multiplier = 10000
        self.__ratedict = dict(self.__termsrates)

    @property
    def fxrate(self):
        return self.__fxrate

    @fxrate.setter
    def fxrate(self, fxrate):
        try:
            self.__fxrate = float(fxrate)
        except:
            raise Exception('fxrate must be a number')

    @property
    def current_date(self):
        return self.__current_date

    @current_date.setter
    def current_date(self, current_date):
        if current_date is not None:
            try:
                self.__current_date = dt64(current_date)
            except:
                raise ValueError('current date is not a date')
        else:
            self.__current_date = current_date

    @property
    def money(self):
        return self.__money

    @money.setter
    def money(self, money):
        if isinstance(money, STRt):
            self.__money = money
        else:
            self.__money = None

    @property
    def base(self):
        return self.__base

    @base.setter
    def base(self, base):
        if isinstance(base, STRt):
            self.__base = base
        else:
            self.__base = None

    @property
    def multiplier(self):
        return self.__multiplier

    @multiplier.setter
    def multiplier(self, multiplier):
        try:
            self.__multiplier = float(multiplier)
        except:
            raise Exception('multiplier is not a number')


    @property
    def swappoint(self, swappoint):
        return self.__swappoint

    @swappoint.setter
    def swappoint(self, swappoint):
        if isinstance(swappoint, dict):
            tenors = []
            therates = []
            for k in swappoint:
                if swappoint.get(k) is not None and self.__termsorder.get(k) is not None:
                    self.__ratedict[k] = swappoint[k]
                    tenors.append(k)
                    therates.append(swappoint[k])
            self.__data = {'tenors': tenors, 'rates': therates}

        elif isinstance(swappoint, Data):
            self.__data = swappoint.data
        else:
            self.__data = {}

    @property
    def start_basis(self):
        return self.__start_basis

    @start_basis.setter
    def start_basis(self, start_basis):

        if start_basis is not None:
            if start_basis in self.__spotbasis:
                self.__start_basis = start_basis
            else:
                raise ValueError('start_basis is not in the list of acceptable values')
        else:
            self.__start_basis = 'Spot'

    @property
    def business_day(self):
        return self.__business_day

    @business_day.setter
    def business_day(self, business_day):
        if business_day is not None:
            if business_day in business_days:
                self.__business_day = business_day
            else:
                raise ValueError('business_day is not in the list of acceptable values')
        else:
            self._business_day = 'Modified Following'


    def calc_impliedswap(self, holidays = []):
        mclass = self.__money
        bclass = self.__base

        moneydf = mclass.calcdf()
        basedf = bclass.calcdf()

        mcurve = _getcurve_dict(moneydf)
        bcurve = _getcurve_dict(basedf)


        tenors =[]
        therates =[]
        for k in self.__ratedict:
            tenors.append(k)
            therates.append(self.__ratedict[k])

        tomdate, spotdate, spotnext = _calc_ondate(self.current_date, holidays=holidays)
        swap_arr = []
        for k in self.__ratedict:
            swap_dict = {}
            swap_dict['tenor'] = k
            swap_dict['point'] = self.__ratedict[k]

            if swap_dict['tenor'] == 'O/N':
                swap_dict['start'] = self.__current_date
                swap_dict['maturity'] = tomdate
                swap_dict['stime'] = 0
            elif swap_dict['tenor'] == 'T/N':
                swap_dict['start'] = tomdate
                swap_dict['maturity'] = spotdate
                swap_dict['stime'] = float(day_cf('Actual/365', self.__current_date, tomdate))
            elif swap_dict['tenor'] == 'S/N':
                swap_dict['start'] = spotdate
                swap_dict['maturity'] = spotnext
                swap_dict['stime'] = float(day_cf('Actual/365', self.__current_date, spotdate))
            else:
                if self.__start_basis == 'Same Day':
                    swap_dict['start'] = self.__current_date
                    swap_dict['stime'] = 0
                elif self.__start_basis == 'Tom':
                    swap_dict['start'] = tomdate
                    swap_dict['stime'] = float(day_cf('Actual/365', self.__current_date, tomdate))
                elif self.__start_basis == 'Spot':
                    swap_dict['start'] = spotdate
                    swap_dict['stime'] = float(day_cf('Actual/365', self.__current_date, spotdate))
                swap_dict['maturity'] = ttm(swap_dict['start'], swap_dict['tenor'], 'Actual/365', self.__business_day)

            stime = float(day_cf('Actual/365', self.__current_date, swap_dict['start']))
            time = float(day_cf('Actual/365', self.__current_date, swap_dict['maturity']))

            swap_dict['time'] = float(day_cf('Actual/365', self.__current_date, swap_dict['maturity']))
            mon_df = interpolation(mcurve['time'], mcurve['df'], swap_dict['time'])/interpolation(mcurve['time'], mcurve['df'], swap_dict['stime'])
            mon_dcf = day_cf(mclass.day_count, swap_dict['start'], swap_dict['maturity'] )
            swap_dict['mrate'] = (1/mon_df - 1) / (mon_dcf * 0.01)

            bas_df = interpolation(bcurve['time'], bcurve['df'], swap_dict['time'])/interpolation(bcurve['time'], bcurve['df'], swap_dict['stime'])
            bas_dcf = day_cf(bclass.day_count, swap_dict['start'], swap_dict['maturity'] )
            swap_dict['brate'] = (1/bas_df - 1) / (bas_dcf * 0.01)

            try:
                spt = float(swap_dict['point'])
                swap_dict['imrate']= _calc_impliedmoney(self.__fxrate, spt/ self.__multiplier, bas_df, mon_dcf)
                swap_dict['ibrate']= _calc_impliedbase(self.__fxrate, spt/ self.__multiplier, mon_df, bas_dcf)
            except:
                swap_dict['imrate'] = None
                swap_dict['ibrate']= None
            forward = self.__fxrate * bas_df/mon_df
            point = (forward - self.__fxrate) * self.__multiplier
            swap_dict['impoint'] = point



            swap_arr.append(swap_dict)

        return swap_arr


@njit('float64(float64, float64, float64, float64)')
def _calc_impliedmoney(fxrate, swappoint, basedf, moneydcf):
    # fwdrate = fxrate * basedf/mon_df
    # mondf = (fxrate * basedf)/fwdrate
    # 1 / (1 + monrate * moneydcf * 0.01) = (fxrate * basedf)/fwdrate
    # (1 + monrate * moneydcf * 0.01) = fwdrate/(fxrate * basedf)
    fwdrate = fxrate + swappoint
    return (fwdrate/(fxrate * basedf) - 1)/ ( moneydcf * 0.01)


@njit('float64(float64, float64, float64, float64)')
def _calc_impliedbase(fxrate, swappoint, moneydf, basedcf):
    # fwdrate = fxrate * basedf / mon_df
    # basedf = fwdrate * mondf / fxrate
    # 1 / (1 + baserate * basedcf * 0.01) = fwdrate * mondf / fxrate
    # (1 + baserate * basedcf * 0.01) = fxrate / (fwdrate * mondf)
    fwdrate = fxrate + swappoint
    return (fxrate / (fwdrate * moneydf) - 1 ) / (basedcf * 0.01)


def _calc_impliedswap(mcurve, bcurve, starttime, endtime, fxrate):
    basedf = list(bcurve['df'])
    basedf.insert(0,1)
    basetime = list(bcurve['time'])
    basetime.insert(0,0)

    moneydf = list(mcurve['df'])
    moneydf.insert(0,1)
    moneytime = list(mcurve['time'])
    moneytime.insert(0,0)

    points = []
    if isinstance(starttime,list):
        base_startdf = interpolation(basetime, basedf, starttime, model='chip', method=None )
        base_enddf = interpolation(basetime, basedf, endtime, model='chip', method=None )
        df_base = list(map(lambda start, end: end/start, base_startdf, base_enddf))

        money_startdf = interpolation(moneytime, moneydf, starttime, model='chip', method=None )
        money_enddf = interpolation(moneytime, money_startdf, endtime, model='chip', method=None )
        df_money = list(map(lambda start, end: end/start, money_startdf, money_enddf))

        forwards = list(map(lambda bdf,mdf: fxrate * bdf / mdf, df_base, df_money))
        points = list(map(lambda fwd: fwd - fxrate, forwards))
    return points

def _getcurve_dict(data_array):
    """
    An internal function to convert array of dictionary (eg [{'time': value0, 'df_tocdate': value1},...]) to dictionary containing array (eg {'time': [], 'df': []}). Used to converts arrays of discount factors to dictionary of discount factors. Source must have the following keys:
    i. time
    ii. df_tocdate
    Returns:
    ========
    dict: {'time':[], 'df': []}
    """
    datalen = len(data_array)
    time = []
    df =[]
    for i in range(datalen):
        datum = data_array[i]
        time.append(datum['time'])
        df.append(datum['df_tocdate'])
    time.insert(0,0)
    df.insert(0,1)

    return {'time': time, 'df': df}


def _calc_maturities(start_basis, day_count, bus_day, dates, tenors):
    if start_basis == 'Same Day':
        return ttm(dates['current'], tenors, day_count,
                         bus_day)
    elif start_basis == 'Tom':
        return ttm(dates['Tom'], tenors, day_count,
                         bus_day)
    elif start_basis == 'Spot':
        return ttm(dates['Spot'], tenors, day_count,
                         bus_day)


def _calc_ondate(current_date, holidays =[]):
    bdc = cal(weekmask='1111100', holidays=holidays)
    tomdate = bus_off(current_date, 1, roll='forward', busdaycal=bdc)
    spotdate = bus_off(tomdate, 1, roll='forward', busdaycal=bdc)
    spotnext = bus_off(spotdate, 1, roll='forward', busdaycal=bdc)
    return (tomdate,spotdate,spotnext)

# %%
def _calc_shorttenor(tenors, rates, setting, holidays=[]):

    v_date = dt64(setting['cdate'])
    tenors1 = tenors[3:]
    tenors2 = tenors[:3]
    tomdate, spotdate, spotnext = _calc_ondate(v_date, holidays=holidays)
    dates = {'current': v_date, 'Tom': tomdate, 'Spot': spotdate}
    data_len = len(tenors)
    data = []
    rates1 = rates[3:]
    rates2 = rates[:3]

    ratebasis = 0
    if setting['rate_basis'] == 'Discount Rate':
        ratebasis = 1

    data_len = len(tenors2)
    for i in range(data_len):
        datum = {}
        datum['tenor'] = tenors2[i]
        datum['rate'] = rates2[i]
        if datum['tenor'] == 'O/N':
            datum['start'] = v_date
            datum['maturity'] = tomdate
        elif datum['tenor'] == 'T/N':
            datum['start'] = tomdate
            datum['maturity'] = spotdate
        elif datum['tenor'] == 'S/N':
            datum['start'] = spotdate
            datum['maturity'] = spotnext

        datum['stime'] = float(day_cf('Actual/365', v_date,
                                    datum['start']))
        datum['dcf'] = float(day_cf(setting['day_count'], datum['start'],
                                    datum['maturity']))
        if datum['rate'] is not None:
            datum['df'] = _strate_to_df(datum['rate'], datum['dcf'], ratebasis)
        else:
            datum['df'] = None
        datum['time'] = day_cf('Actual/365', v_date, datum['maturity'])
        data.append(datum)

    maturities = _calc_maturities(setting['start_basis'], setting['day_count'], setting['bus_day'], dates, tenors1)
    start_date = None
    if setting['start_basis'] == 'Same Day':
        start_date = v_date
    elif setting['start_basis'] == 'Tom':
        start_date = tomdate
    elif setting['start_basis'] == 'Spot':
        start_date = spotdate

    stime = day_cf('Actual/365', v_date, start_date)
    data_len = len(tenors1)
    for i in range(data_len):
        datum = {}
        datum['tenor'] = tenors1[i]
        datum['rate'] = rates1[i]
        datum['maturity'] = maturities[i]
        datum['start'] = start_date
        datum['stime'] = stime
        datum['dcf'] = float(day_cf(setting['day_count'], datum['start'],
                                    datum['maturity']))
        if datum['rate'] is not None:
            datum['df'] = _strate_to_df(datum['rate'], datum['dcf'], ratebasis)
        else:
            datum['df'] = None
        datum['time'] = float(day_cf('Actual/365', datum['start'],
                                    datum['maturity']))
        data.append(datum)

    # Calculate df to current date
    data_len = len(data)
    if setting['start_basis'] == 'Same Day':
        for i in range(data_len):
            datum = data[i]
            if datum['tenor'] == 'O/N':
                if datum.get('df') is not None:
                    datum['df_tocdate'] = datum['df']
                else:
                    datum['df_tocdate'] = None
            elif datum['tenor'] == 'T/N' and datum.get('df') is not None:
                on = dict(data[i-1])
                if on.get('df') is not None:
                    datum['df_tocdate'] = on['df'] * datum['df']
                else:
                    on['rate'] = datum['rate']
                    on['df'] = 1 / (1 + on['rate'] * 0.01 * on['dcf'])
                    datum['df_tocdate'] = on['df'] * datum['df']

            elif datum['tenor'] == 'S/N' and datum.get('df') is not None:
                tn = dict(data[i-1])
                if tn.get('df_tocdate') is not None:
                    datum['df_tocdate'] = tn['df_tocdate'] * datum['df']
                else:
                    on = dict(data[i-2])
                    on['rate'] = datum['rate']
                    on['df'] = 1 / (1 + on['rate'] * 0.01 * on['dcf'])
                    on['df_tocdate'] = on['df']
                    tn['rate'] = datum['rate']
                    tn['df'] = 1 / (1 + tn['rate'] * 0.01 * tn['dcf'])
                    tn['df_tocdate'] = on['df'] * tn['df']
                    datum['df_tocdate'] = tn['df'] * datum['df']
            else:
                datum['df_tocdate'] = datum['df']

    elif setting['start_basis'] == 'Spot':
        _adjustdf_fromspot(data)

    elif setting['start_basis'] == 'Tom':
        df_totom = None
        ondf = None
        ontime = None

        for i in range(0, 1, 1):
            datum = data[i]

            if datum['tenor'] == 'O/N':
                ondf = datum['df']
                ontime = datum['time']
                if ondf is not None:
                    datum['df_tocdate'] = ondf

        tomtime = ontime
        if ondf is None:
            df_totom = ondf
            # get discount factor to interpolate
            x_axis = [0]
            y_axis = [1]
            for n in range(1, data_len, 1):
                datum1 = data[n]
                if datum1['df'] is not None:
                    x_axis.append(datum1['time'])
                    y_axis.append(datum1['df'])

            if ondf is None:
                x_value = tomtime
                df_totom = interpolation(x_axis, y_axis, x_value)


        for i in range(1, data_len, 1):
            datum = data[i]
            if datum['df'] is not None:
                datum['df_tocdate'] = datum['df'] * df_totom

    return data



def _adjustdf_fromspot(data):
    df_tospot = None
    ondf = None
    tndf = None
    ontime = None
    tntime = None
    data_len = len(data)
    for i in range(0, 2, 1):
        datum = data[i]
        if datum['tenor'] == 'O/N':
            ondf = datum['df']
            if ondf is not None:
                datum['df_tocdate'] = ondf
            ontime = datum['time']
        if datum['tenor'] == 'T/N':
            tndf = datum['df']
            tntime = datum['time']
            if ondf is not None and tndf is not None:
                datum['df_tocdate'] = ondf * tndf


    spottime = tntime
    if ondf is not None and tndf is not None:
        df_tospot = ondf * tndf
    else:
        # get discount factor to interpolate
        x_axis = [0]
        y_axis = [1]
        for n in range(2, data_len, 1):
            if datum1['df'] is not None:
                x_axis.append(datum1['time'])
                y_axis.append(datum1['df'])

        if ondf is None and tndf is None:
            x_value = spottime
            df_tospot = interpolation(x_axis, y_axis, x_value)

        elif ondf is None:
            x_value = ontime
            ondf = interpolation(x_axis, y_axis, x_value)
            df_tospot = ondf * tndf

        elif tndf is None:
            x_value = tntime - ontime
            tndf = interpolation(x_axis, y_axis, x_value)
            df_tospot = ondf * tndf

    for i in range(2, data_len, 1):
        datum = data[i]

        if datum['df'] is not None:
            datum['df_tocdate'] = datum['df'] * df_tospot
            if datum['tenor'] != 'S/N':
                datum['time'] = datum['time'] + datum['stime']


@njit('float64(float64, float64, int16)')
def _strate_to_df(rate, dcf, ratebasis):
    # Discount Rate
    if ratebasis == 1:
        return (1 - rate * 0.01 * dcf)
    # Money Market
    else:
        return (1 / (1 + rate * 0.01 * dcf))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 21:02:58 2020

@author: RMS671214
"""

from interestrate import fixincome as fi
import numpy as np
from interestrate import rmp_dates as rd
from interestrate import discount_curve as dcurve
from interestrate import rmp_curves as rcurve
import pandas as pd
# %%

myfl = {}
myfl['issue_date'] = np.datetime64('2018-10-22')
myfl['value_date'] = np.datetime64('2021-10-22')
myfl['maturity'] = np.datetime64('2028-10-22')
myfl['day_count'] = 'Actual/365 Fixed'
myfl['frequency'] = 'Semi-Annual'
myfl['business_day'] = 'No Adjustment'
myfl['date_generation'] = rd.date_gen_method[0]
myfl['face_value'] = 1000000
myfl['current_coupon'] = 3.65
myfl["principal_exchange"] = False
myfl["fixing_basis"] = "Same Day"
myfl["margin"] = 0.00


# %%

rate = {"value_date": str(myfl['value_date']),
        "st_busday": "Modified Following",
        "st_ratebasis": "Money Market", "st_daycount": "Actual/365",
        "lt_busday": "No Adjustment", "lt_frequency": "Semi-Annual",
        "lt_daycount": "Actual/Actual",
        "rates": {'O/N': 2.30, '1W': 2.35, '1M': 2.45, '3M': 2.55,
                  '6M': 2.65, '12M': 2.75, '1Y': 2.70, '2Y': 2.80, '3Y': 2.90,
                  '5Y': 3.00, '10Y': 3.10, '30Y': 3.25}}


dfs = dcurve.discount_factor_gen(rate, return_type="times")
x_axis = [x["times"] for x in dfs]
y_axis = [x["df"] for x in dfs]
df_func = rcurve.interpolation(x_axis, y_axis, float(1/366), is_function=True)
pd_ori = pd.DataFrame(dfs)

# %%
results = fi.floatleg(myfl, rate, val_curve=rate)
#print("FLOAT LEG STRUCTURE")
#print("=================")
#print(results["structure"])

print("FLOAT LEG RIKS")
print("==============")
print(results["risks"])

pd_struc = pd.DataFrame(results["structure"])


# %%
